import { __awaiter, __decorate } from "tslib";
import { Connect, dataServerResultFilter, Datasource, DatasourceDefaults, FieldTypeEnum, MessageType, normaliseCriteria, ResourceType, } from '@genesislcap/foundation-comms';
import { LifecycleMixin } from '@genesislcap/foundation-utils';
import { attr, customElement, DOM, observable } from '@microsoft/fast-element';
import { FoundationElement } from '@microsoft/fast-foundation';
import { BehaviorSubject } from 'rxjs';
import { debounceTime, skip, tap } from 'rxjs/operators';
import { dateTimeValueFormatter, dateValueFormatter, getFilterParamsByFieldType, } from '../grid-pro.definitions';
import { GridProRendererTypes } from '../grid-pro.types';
import { logger } from '../utils/logger';
const criteriaDelimiter = ';';
const criteriaJoin = ' && ';
const withoutFullInit = null;
/**
 * The Genesis Datasource element, for CSRM-compatible data fetching and used exclusively by the GridPro element.
 * @remarks Only supports Client-Side Row Model.
 * @public
 */
let GridProGenesisDatasource = class GridProGenesisDatasource extends LifecycleMixin(FoundationElement) {
    constructor() {
        super(...arguments);
        this.isSnapshot = false;
        this.maxRows = DatasourceDefaults.MAX_ROWS_250;
        this.maxView = DatasourceDefaults.MAX_VIEW_1000;
        this.pollingInterval = DatasourceDefaults.REQ_REP_POLLING_INTERVAL_MS;
        this.movingView = false;
        this.reverse = false;
        this.restartOnReconnection = false;
        this.isRequestServer = false;
        this.requiresFullRowDataAndColDefs = true;
        this.dataSubWasLoggedOff = false;
        this.rows = new Map();
        this.criteriaFromFilters = new Map();
        this.update = new BehaviorSubject(new Map());
    }
    criteriaChanged(oldCriteria, newCriteria) {
        const criteriaIsNotDuplicate = oldCriteria !== normaliseCriteria(newCriteria, criteriaDelimiter);
        if (this.datasource.initialized && criteriaIsNotDuplicate) {
            if (this.dataSub) {
                this.dataSub.unsubscribe();
                this.dataSub = undefined;
            }
            this.loadResourceData(withoutFullInit);
        }
    }
    resourceNameChanged(oldValue, newValue) {
        if (!oldValue || oldValue === newValue)
            return;
        this.restart();
    }
    connectedCallback() {
        super.connectedCallback();
        const shouldRunConnect = this.shouldRunConnect;
        DOM.queueUpdate(() => {
            if (!shouldRunConnect)
                return;
            this.init();
        });
        this.connectionSub = this.connect.isConnected$.subscribe((isConnected) => {
            if (isConnected && this.agGrid.attributes['ds-disconnected']) {
                this.agGrid.removeAttribute('ds-disconnected');
                if (this.restartOnReconnection) {
                    this.restart();
                }
            }
            if (!isConnected) {
                this.agGrid.setAttribute('ds-disconnected', 'ds-disconnected');
            }
        });
    }
    disconnectedCallback() {
        super.disconnectedCallback();
        const shouldRunDisconnect = this.shouldRunDisconnect;
        DOM.queueUpdate(() => {
            if (!shouldRunDisconnect)
                return;
            this.deinit();
        });
    }
    deepClone() {
        const copy = super.deepClone();
        copy.deferredGridOptions = structuredClone(this.deferredGridOptions);
        copy.request = structuredClone(this.request);
        return copy;
    }
    get agGrid() {
        return this.parentElement;
    }
    // Genesis Specific
    get rowId() {
        if (this.rowIdAttr)
            return this.rowIdAttr; // Allow user to override
        if (this.isRequestServer)
            return 'RECORD_ID'; // REQUEST_SERVER default record id
        return 'ROW_REF'; // DATASERVER default record id
    }
    /**
     * Initializes the datasource.
     * @public
     * @remarks This method is called automatically when the element is connected to the DOM.
     */
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.agGrid &&
                this.datasource.validResourceName(this.resourceName) &&
                this.connect.isConnected) {
                const meta = yield this.connect.getMetadata(this.resourceName);
                this.isRequestServer = meta.TYPE === ResourceType.REQUEST_SERVER;
                const gridOptions = Object.assign({ getRowId: (params) => params.data[this.rowId] }, this.deferredGridOptions);
                this.agGrid.gridOptions = Object.assign({}, gridOptions);
                this.agGrid.addEventListener('onGridReady', () => __awaiter(this, void 0, void 0, function* () {
                    yield this.loadResourceData();
                }), { once: true });
                const filterDebounceTime = 600;
                this.updateSub = this.update
                    .pipe(skip(1), debounceTime(filterDebounceTime), tap((f) => logger.debug('filters (debounced): ', f)))
                    .subscribe((value) => {
                    this.loadResourceData(withoutFullInit);
                });
                return;
            }
            logger.warn(`Application not connected, falling back to local columnDefs/rowData`);
        });
    }
    /**
     * Deinitialises the datasource, resetting it to its initial state.
     * @public
     */
    deinit() {
        this.reset();
    }
    /**
     * Resets the datasource to its initial state.
     * @public
     * @deprecated Use `deinit` instead
     */
    reset() {
        if (this.dataSub) {
            this.dataSub.unsubscribe();
            this.dataSub = undefined;
        }
        if (this.updateSub) {
            this.updateSub.unsubscribe();
            this.updateSub = undefined;
        }
        if (this.connectionSub) {
            this.connectionSub.unsubscribe();
            this.connectionSub = undefined;
        }
        this.isRequestServer = false;
        this.requiresFullRowDataAndColDefs = true;
        this.dataSubWasLoggedOff = false;
        this.dataLogoff();
        this.clearRowData();
        this.criteriaFromFilters = new Map();
        this.update = new BehaviorSubject(new Map());
        this.datasource.deinit();
    }
    /**
     * Restarts the datasource, uses `deinit` and `init` in sequence.
     * @public
     */
    restart() {
        this.deinit();
        this.init();
    }
    /**
     * Force the grid to redispatch the current rows
     */
    refreshRows() {
        this.updateRowData(Array.from(this.rows.values()));
    }
    /**
     * Sends a DATA_LOGOFF when the resource is an active stream
     * @remarks Will only happen for streaming DATASERER resources.
     * @internal
     */
    dataLogoff() {
        if (this.sourceRef && !this.isRequestServer && !this.isSnapshot) {
            this.connect.dataLogoff(this.sourceRef);
        }
        this.sourceRef = undefined;
    }
    clearRowData() {
        var _a, _b;
        this.rows = new Map();
        this.agTransaction = undefined;
        (_b = (_a = this.agGrid) === null || _a === void 0 ? void 0 : _a.gridApi) === null || _b === void 0 ? void 0 : _b.setRowData([]);
    }
    updateRowData(rowData) {
        const agColumnDefs = this.getAgColumnDefs(this.datasource.fieldMetadata);
        if (agColumnDefs) {
            this.agGrid.gridApi.setColumnDefs(agColumnDefs);
        }
        const data = this.rowDataMapper ? this.rowDataMapper(rowData) : rowData;
        this.agGrid.gridApi.setRowData(data);
        this.requiresFullRowDataAndColDefs = false;
    }
    loadResourceData(withFullInit = true) {
        return __awaiter(this, void 0, void 0, function* () {
            const requiresMetadataFetch = withFullInit || !this.datasource.initialized;
            if (!requiresMetadataFetch) {
                this.dataLogoff();
                this.clearRowData();
            }
            const initOK = yield this.datasource.init(this.datasourceOptions(), requiresMetadataFetch);
            if (!initOK) {
                logger.debug(`Genesis Datasource init failed for ${this.resourceName}`);
                this.clearRowData();
                return;
            }
            if (this.isSnapshot) {
                const result = yield this.datasource.snapshot();
                if (result) {
                    const rowData = this.handleSnapshot(result);
                    this.updateRowData(rowData);
                }
                return;
            }
            logger.debug(`requesting stream for ${this.resourceName}`);
            this.dataSub = this.datasource.stream.subscribe((result) => {
                this.sourceRef = result.SOURCE_REF;
                const messageType = result.MESSAGE_TYPE;
                if (messageType && messageType === MessageType.LOGOFF_ACK) {
                    this.dataSubWasLoggedOff = true;
                    return;
                }
                this.dataSubWasLoggedOff = false;
                if (result.ROW) {
                    const nextMessage = dataServerResultFilter(result);
                    this.handleStreamResult(nextMessage);
                }
                else {
                    this.handleStreamResult(result);
                }
            });
        });
    }
    handleSnapshot(result) {
        if (Array.isArray(result))
            return result;
        if (result.REPLY) {
            return result.REPLY;
        }
        if (result.ROW) {
            return dataServerResultFilter(result).inserts;
        }
        return result.inserts;
    }
    datasourceOptions() {
        return {
            criteria: this.buildCriteria(),
            fields: this.fields,
            isSnapshot: this.isSnapshot,
            orderBy: this.orderBy,
            resourceName: this.resourceName,
            request: this.request,
            reverse: this.reverse,
            maxRows: this.maxRows,
            maxView: this.maxView,
            pollingInterval: this.pollingInterval,
            movingView: this.movingView,
        };
    }
    handleStreamResult(result) {
        if (!result)
            return;
        if (this.isRequestServer) {
            this.applyRequestServerData(result);
        }
        else {
            this.applyDataserverData(result);
        }
    }
    applyRequestServerData(requestServerResult) {
        var _a;
        const requestServerData = requestServerResult.REPLY;
        if (!Array.isArray(requestServerData) ||
            !((_a = requestServerResult.MESSAGE_TYPE) === null || _a === void 0 ? void 0 : _a.startsWith('REP_'))) {
            logger.error('received invalid RequestServerResult', requestServerResult);
            return;
        }
        if (this.requiresFullRowDataAndColDefs) {
            requestServerData === null || requestServerData === void 0 ? void 0 : requestServerData.forEach((insertData) => {
                this.rows.set(insertData[this.rowId], insertData);
            });
            const rowData = this.rows.size > 0 ? Array.from(this.rows.values()) : requestServerData;
            this.updateRowData(rowData);
            return;
        }
        this.agTransaction = { add: [], remove: [], update: [] };
        const rowsToUpdate = [];
        const rowsToDelete = Array.from(this.rows.values()).filter((row) => {
            const match = requestServerData.find((data) => data[this.rowId] === row[this.rowId]);
            if (match) {
                rowsToUpdate.push(match);
                return false; // don't delete rows that need to be updated
            }
            return true; // delete rows that don't exist in requestServerData
        });
        const rowsToInsert = requestServerData.filter((data) => !this.rows.has(data[this.rowId]));
        this.handleStreamInserts(rowsToInsert);
        this.handleStreamDeletes(rowsToDelete);
        this.handleStreamUpdates(rowsToUpdate);
        this.agGrid.gridApi.applyTransaction(this.mapTransaction(this.agTransaction));
    }
    applyDataserverData(dataServerResult) {
        var _a;
        if (this.requiresFullRowDataAndColDefs) {
            (_a = dataServerResult.inserts) === null || _a === void 0 ? void 0 : _a.forEach((insertData) => {
                this.rows.set(insertData[this.rowId], insertData);
            });
            const rowData = Array.from(this.rows.values());
            this.updateRowData(rowData);
            return;
        }
        this.agTransaction = { add: [], remove: [], update: [] };
        this.handleStreamInserts(dataServerResult.inserts);
        this.handleStreamDeletes(dataServerResult.deletes);
        this.handleStreamUpdates(dataServerResult.updates);
        this.agGrid.gridApi.applyTransaction(this.mapTransaction(this.agTransaction));
    }
    mapTransaction(transaction) {
        if (!this.rowDataMapper) {
            return transaction;
        }
        return Object.assign(Object.assign({}, transaction), { add: this.rowDataMapper(transaction.add, this.rows), update: this.rowDataMapper(transaction.update, this.rows), remove: this.rowDataMapper(transaction.remove, this.rows) });
    }
    handleStreamInserts(insertedRows) {
        insertedRows === null || insertedRows === void 0 ? void 0 : insertedRows.forEach((insertData) => {
            if (this.rows.has(insertData[this.rowId]) || this.dataSubWasLoggedOff) {
                const rowToBeUpdated = this.rows.get(insertData[this.rowId]);
                const updatedRow = Object.assign(Object.assign({}, rowToBeUpdated), insertData);
                this.agTransaction.update.push(updatedRow);
            }
            else {
                this.agTransaction.add.push(insertData);
                this.agTransaction.addIndex = 0;
            }
            this.rows.set(insertData[this.rowId], insertData);
        });
    }
    handleStreamDeletes(deletedRows) {
        deletedRows === null || deletedRows === void 0 ? void 0 : deletedRows.forEach((deleteData) => {
            this.agTransaction.remove.push({ [this.rowId]: deleteData[this.rowId] });
            this.rows.delete(deleteData[this.rowId]);
        });
    }
    handleStreamUpdates(updatedRows) {
        updatedRows === null || updatedRows === void 0 ? void 0 : updatedRows.forEach((updateData) => {
            const rowToBeUpdated = this.rows.get(updateData[this.rowId]);
            const updatedRow = Object.assign(Object.assign({}, rowToBeUpdated), updateData);
            this.agTransaction.update.push(updatedRow);
            this.rows.set(updateData[this.rowId], updatedRow);
        });
    }
    // FUTURE: Work with the field types vs ag grid column def types!
    getAgColumnDefs(fieldsMetadata) {
        const colDefsFromGenesisData = fieldsMetadata === null || fieldsMetadata === void 0 ? void 0 : fieldsMetadata.flatMap((field) => {
            var _a;
            const cellConfigs = {};
            cellConfigs.filterParams = getFilterParamsByFieldType(field.type);
            switch (field.type) {
                case FieldTypeEnum.BOOLEAN:
                    if (this.agGrid.autoCellRendererByType) {
                        cellConfigs.cellRenderer = GridProRendererTypes.boolean;
                    }
                    break;
                case FieldTypeEnum.DATE:
                    cellConfigs.valueFormatter = dateValueFormatter;
                    break;
                case FieldTypeEnum.DATETIME:
                    cellConfigs.valueFormatter = dateTimeValueFormatter;
                    break;
                default:
                    break;
            }
            const templateColDef = (_a = this.agGrid.gridApi) === null || _a === void 0 ? void 0 : _a.getColumnDef(field.name);
            if (this.agGrid.onlyTemplateColDefs && !templateColDef) {
                return [];
            }
            if (this.fields && !this.fields.includes(field.name)) {
                return [];
            }
            return Object.assign(Object.assign({ field: field.name }, cellConfigs), templateColDef);
        });
        const colDefsMergedWithTemplateDefs = this.agGrid.applyTemplateDefinitions(colDefsFromGenesisData);
        return colDefsMergedWithTemplateDefs;
    }
    buildCriteria() {
        var _a;
        const initialCriteria = ((_a = this.criteria) === null || _a === void 0 ? void 0 : _a.split(criteriaDelimiter)) || [];
        const criteria = initialCriteria.concat(Array.from(this.criteriaFromFilters.values()));
        return normaliseCriteria(criteria.join(criteriaJoin), criteriaDelimiter);
    }
    setFilter(fieldName, newFilter) {
        this.criteriaFromFilters.set(fieldName, newFilter);
        this.update.next(this.criteriaFromFilters);
    }
    removeFilter(fieldName) {
        this.criteriaFromFilters.delete(fieldName);
        this.update.next(this.criteriaFromFilters);
    }
};
__decorate([
    Connect
], GridProGenesisDatasource.prototype, "connect", void 0);
__decorate([
    Datasource
], GridProGenesisDatasource.prototype, "datasource", void 0);
__decorate([
    observable
], GridProGenesisDatasource.prototype, "deferredGridOptions", void 0);
__decorate([
    attr
], GridProGenesisDatasource.prototype, "criteria", void 0);
__decorate([
    attr
], GridProGenesisDatasource.prototype, "fields", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'is-snapshot' })
], GridProGenesisDatasource.prototype, "isSnapshot", void 0);
__decorate([
    attr({ attribute: 'max-rows' })
], GridProGenesisDatasource.prototype, "maxRows", void 0);
__decorate([
    attr({ attribute: 'max-view' })
], GridProGenesisDatasource.prototype, "maxView", void 0);
__decorate([
    attr({ attribute: 'polling-interval' })
], GridProGenesisDatasource.prototype, "pollingInterval", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'moving-view' })
], GridProGenesisDatasource.prototype, "movingView", void 0);
__decorate([
    attr({ attribute: 'order-by' })
], GridProGenesisDatasource.prototype, "orderBy", void 0);
__decorate([
    observable
], GridProGenesisDatasource.prototype, "request", void 0);
__decorate([
    attr({ attribute: 'resource-name' })
], GridProGenesisDatasource.prototype, "resourceName", void 0);
__decorate([
    attr({ mode: 'boolean' })
], GridProGenesisDatasource.prototype, "reverse", void 0);
__decorate([
    attr({ attribute: 'row-id' })
], GridProGenesisDatasource.prototype, "rowIdAttr", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'restart-on-reconnection' })
], GridProGenesisDatasource.prototype, "restartOnReconnection", void 0);
__decorate([
    observable
], GridProGenesisDatasource.prototype, "rowDataMapper", void 0);
GridProGenesisDatasource = __decorate([
    customElement({
        name: 'grid-pro-genesis-datasource',
    })
], GridProGenesisDatasource);
export { GridProGenesisDatasource };
